#
# Authors: Ma He <mahe.itsec@gmail.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import hashlib
import logging
import socket
import sys

import dns.resolver
import requests
from silentdune_client.builders import iptables as ipt
from silentdune_client.models.remote_config import RemoteConfig
from silentdune_client.modules import BaseModule, QueueTask
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, TASK_FIREWALL_INSERT_RULES, \
    TASK_FIREWALL_DELETE_SLOT
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_egress_ingress_rule, \
    create_iptables_remote_config, create_iptables_egress_rule_dest, create_iptables_ingress_rule_source
from silentdune_client.modules.firewall.manager.slots import Slots

_logger = logging.getLogger('sd-client')

module_list = {
    'Silent Dune Remote Config': {
        'module': 'SilentDuneRemoteConfigModule',
    },
}


class SilentDuneRemoteConfigModule(BaseModule):
    """Silent Dune Remote Config Module"""

    _t_check_remote_config = 0  # Timed event property for debug message loop.
    _t_reload_task = 0  # Timed event property for reloading firewall rules.

    _remote_config_url = None
    _remote_config_dns_name = None
    _remote_config_check_interval = 0

    _remote_config_hash = None

    _slot_config_access = Slots.rc_access
    _slot_config_apply = Slots.rc_apply

    _startup = True

    def __init__(self):

        self._arg_name = 'remoteconfig'  # Set argparser name
        self._config_section_name = 'remote_config'  # Set configuration file section name

        # Enable multi-threading
        self.wants_processing_thread = True

        self._enabled = False
        self._remote_config_check_interval = 4 * 3600  # 4 hours

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('remote config module', 'Silent Dune Remote Config module')

        group.add_argument('--remoteconfig-mod-enable', action='store_true',  # noqa
                           help=_('Enable the remote config module'))  # noqa

        group.add_argument(_('--remote-config-url'), help=_('Set remote config url'),  # noqa
                           default=None, type=str)  # noqa
        group.add_argument(_('--remote-config-dns-name'), help=_('Set dns name for remote config'),  # noqa
                           default=None, type=str)  # noqa
        group.add_argument(_('--remote-config-check-interval'), help=_('Set check interval for remote config'),  # noqa
                           default=4 * 3600, type=str)  # noqa

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """

        # Check for conflicting arguments.
        if '--remoteconfig-mod-enable' in sys.argv and (
                '--remote-config-url' in sys.argv or
                '--remote-config-dns-name' in sys.argv or
                '--remote-config-check-interval' in sys.argv):
            print(
                'sdc-install: argument --remoteconfig-mod-enable conflicts with other remote config module arguments.')
            return False

        if args.remoteconfig_mod_enable:
            self._enabled = True

        # Check for any required arguments here

        self._remote_config_url = args.remote_config_url
        self._remote_config_dns_name = args.remote_config_dns_name
        self._remote_config_check_interval = args.remote_config_check_interval

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """

        # See if we are enabled or not
        try:
            self._enabled = True if config.get(self._config_section_name, 'enabled').lower() == 'yes' else False
        except:
            _logger.debug('{0} configuration section not found in configuration file.'.format(
                self._config_section_name))
            self._enabled = False

        # Only worry about the rest of the configuration items if we are enabled.
        if self._enabled:
            self._remote_config_url = config.get(self._config_section_name, 'remote_config_url').lower()
            self._remote_config_dns_name = config.get(self._config_section_name, 'remote_config_dns_name').lower()
            self._remote_config_check_interval = int(
                config.get(self._config_section_name, 'remote_config_check_interval'))

        return True

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'enabled', 'yes' if self._enabled else 'no')
        config.set(self._config_section_name, 'remote_config_url',
                   self._remote_config_url if self._remote_config_url else None)
        config.set(self._config_section_name, 'remote_config_dns_name',
                   self._remote_config_dns_name if self._remote_config_dns_name else 'silentdune')
        config.set(self._config_section_name, 'remote_config_check_interval',
                   self._remote_config_check_interval if self._remote_config_check_interval else 4 * 3600)

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune remote config Module Configuration\n'))  # noqa
        config.set_comment(self._config_section_name, 'remote_config_url',
                           _('; Set remote config url.\n'))  # noqa
        config.set_comment(self._config_section_name, 'remote_config_dns_name',
                           _('; Set dns override for remote config.\n'))  # noqa
        config.set_comment(self._config_section_name, 'remote_config_check_interval',
                           _('; Set check interval for remote config.\n'))  # noqa

        return True

    def service_startup(self):
        _logger.debug('{0}: module startup called'.format(self.get_name()))

        if not self.validate_config(self.config):
            _logger.error('{0}: module configuration validation failed.'.format(self.get_name()))
            return False

        if not self._enabled:
            return True

        return True

    def service_shutdown(self):
        _logger.debug('{0}: module shutdown called'.format(self.get_name()))

    def process_loop(self):

        if self._startup:
            self._startup = False
            _logger.debug(
                '{0} remote config debug message {1}, {2}, {3}'.format(self.get_name(), self._remote_config_url,
                                                                       self._remote_config_dns_name,
                                                                       self._remote_config_check_interval))
            self.allow_remote_config_access()
            self.apply_remote_config()

        if self.timed_event('_t_check_remote_config', self._remote_config_check_interval):
            self.apply_remote_config()

    def allow_remote_config_access(self):

        rules = list()

        url = self._remote_config_url
        if 'http://' in url:
            hostname = url.split('http://')[1].split('/')[0]
            rules.append(self.get_http_rule_by_hostname(hostname))
        elif 'https://' in url:
            hostname = url.split('https://')[1].split('/')[0]
            rules.append(self.get_https_rule_by_hostname(hostname))
        elif 'ftp://' in url:
            hostname = url.split('ftp://')[1].split('/')[0]
            rules.append(self.get_ftp_rule_by_hostname(hostname))

        # Notify the firewall module to reload the rules.
        task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                         src_module=self.get_name(),
                         dest_module=SilentDuneClientFirewallModule().get_name(),
                         data=rules)
        self.send_parent_task(task)

    def get_http_rule_by_hostname(self, hostname):

        rules = list()

        if hostname:
            if hostname.find(':') == -1:
                ipaddrs = self.get_ip_from_hostname(hostname, 80)
                for ipaddr in ipaddrs:
                    _logger.debug(
                        '{0}: adding rules for: hostname: {1}, ip addr: {2}, port: {3}'.format(
                            self.get_name(), hostname, ipaddr, 80))
                    rules.append(create_iptables_egress_ingress_rule(ipaddr, 80, 'tcp', self._slot_config_access,
                                                                     transport=ipt.TRANSPORT_AUTO))
            else:
                port = int(hostname.split(':')[1])
                hostname = hostname.split(':')[0]
                ipaddrs = self.get_ip_from_hostname(hostname, port)
                for ipaddr in ipaddrs:
                    _logger.debug(
                        '{0}: adding rules for access of Remote Config hostname: {1}, ip addr: {2}, port: {3}'.format(
                            self.get_name(), hostname, ipaddr, port))
                    rules.append(create_iptables_egress_ingress_rule(ipaddr, port, 'tcp', self._slot_config_access,
                                                                     transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_https_rule_by_hostname(self, hostname):

        rules = list()

        if hostname:
            if hostname.find(':') == -1:
                ipaddrs = self.get_ip_from_hostname(hostname, 443)
                for ipaddr in ipaddrs:
                    _logger.debug(
                        '{0}: adding rules for: hostname: {1}, ip addr: {2}, port: {3}'.format(
                            self.get_name(), hostname, ipaddr, 443))
                    rules.append(create_iptables_egress_ingress_rule(ipaddr, 443, 'tcp', self._slot_config_access,
                                                                     transport=ipt.TRANSPORT_AUTO))
            else:
                port = int(hostname.split(':')[1])
                hostname = hostname.split(':')[0]
                ipaddrs = self.get_ip_from_hostname(hostname, port)
                for ipaddr in ipaddrs:
                    _logger.debug(
                        '{0}: adding rules for: hostname: {1}, ip addr: {2}, port: {3}'.format(
                            self.get_name(), hostname, ipaddr, port))
                    rules.append(create_iptables_egress_ingress_rule(ipaddr, port, 'tcp', self._slot_config_access,
                                                                     transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_ftp_rule_by_hostname(self, hostname):

        rules = list()

        if hostname:
            ipaddrs = self.get_ip_from_hostname(hostname, 21)
            for ipaddr in ipaddrs:
                _logger.debug(
                    '{0}: adding rules for: hostname: {1}, ip addr: {2}, ftp ports: 21, 20, etc'.format(
                        self.get_name(), hostname, ipaddr))
                # FTP control
                rules.append(create_iptables_egress_ingress_rule(ipaddr, 21, u'tcp', self._slot_config_access,
                                                                 transport=ipt.TRANSPORT_AUTO))
                # FTP data transfer
                rules.append(
                    create_iptables_egress_rule_dest(ipaddr, 20, u'tcp', self._slot_config_access, u'ESTABLISHED',
                                                     transport=ipt.TRANSPORT_AUTO))
                rules.append(create_iptables_ingress_rule_source(ipaddr, 20, u'tcp', self._slot_config_access,
                                                                 u'ESTABLISHED,RELATED', transport=ipt.TRANSPORT_AUTO))
                rules.append(create_iptables_egress_rule_dest(ipaddr, None, u'tcp', self._slot_config_access,
                                                              u'ESTABLISHED,RELATED', transport=ipt.TRANSPORT_AUTO))
                rules.append(
                    create_iptables_ingress_rule_source(ipaddr, None, u'tcp', self._slot_config_access, u'ESTABLISHED',
                                                        transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_ip_from_hostname(self, hostname, port):

        ipaddrs = list()

        if hostname:
            if port:
                if 1 < port <= 65536:
                    ais = socket.getaddrinfo(hostname, port, 0, 0, socket.IPPROTO_TCP)
                    for result in ais:
                        ipaddr = result[-1][0]
                        ipaddrs.append(ipaddr)

        return ipaddrs

    def apply_remote_config(self):

        rules = list()
        rq = requests.get(self._remote_config_url)

        if rq.status_code == 200:
            rc = RemoteConfig(rq.json())

            if rc.hash_method == 'sha256':
                hash_value = hashlib.sha256(rq.text).hexdigest()
            else:
                return

            if self.get_dns_txt() == hash_value:

                if self._remote_config_hash != hash_value:

                    if self._remote_config_hash is not None:
                        _logger.debug(
                            '{0}: deleting rules for slot {1}'.format(self.get_name(), self._slot_config_apply))
                        task = QueueTask(TASK_FIREWALL_DELETE_SLOT,
                                         src_module=self.get_name(),
                                         dest_module=SilentDuneClientFirewallModule().get_name(),
                                         data=self._slot_config_apply)
                        self.send_parent_task(task)

                    for rule in rc.rules:
                        print("{0}/{1}, {2}, {3}, {4}, {5}".format(rule.host, rule.mask, rule.port, rule.protocol,
                                                                   rule.uid, rule.gid))
                        _logger.debug('{0}: adding rules slot: {1}, ip address: {2}, port: {3}, protocol: {4}'.format(
                            self.get_name(), self._slot_config_apply, rule.host, rule.port, rule.protocol))
                        rules.append(create_iptables_remote_config(rule.host, rule.mask, rule.port, rule.protocol,
                                                                   self._slot_config_apply, rule.uid, rule.gid,
                                                                   transport=ipt.TRANSPORT_AUTO))

                    if rules:
                        task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                                         src_module=self.get_name(),
                                         dest_module=SilentDuneClientFirewallModule().get_name(),
                                         data=rules)
                        self.send_parent_task(task)

                    self._remote_config_hash = hash_value

    def get_dns_txt(self):

        query = self._remote_config_dns_name + '.entpack.com'
        answers = dns.resolver.query(query, 'TXT')
        for rdata in answers:
            for txt_string in rdata.strings:
                return txt_string

        return None
